@@ -1080,11 +1080,26 @@ var PlacesUtils = {
               // remove tags via the tagging service
               var tags = this._utils.tagging.allTags;
               var uris = [];
+              var bogusTagContainer = false;
               for (let i in tags) {
-                var tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+                var tagURIs = [];
+                // skip empty tags since getURIsForTag would throw
+                if (tags[i])
+                  tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+
+                if (!tagURIs.length) {
+                  // This is a bogus tag container, empty tags should be removed
+                  // automatically, but this does not work if they contain some
+                  // not-uri node, so we remove them manually.
+                  // XXX this is a temporary workaround until we implement
+                  // preventive database maintainance in bug 431558.
+                  bogusTagContainer = true;
+                }
                 for (let j in tagURIs)
                   this._utils.tagging.untagURI(tagURIs[j], [tags[i]]);
               }
+              if (bogusTagContainer)
+                this._utils.bookmarks.removeFolderChildren(rootItemId);
             }
             else if ([this._utils.toolbarFolderId,
                       this._utils.unfiledBookmarksFolderId,
@@ -1168,7 +1183,11 @@ var PlacesUtils = {
         if (aContainer == PlacesUtils.bookmarks.tagsFolder) {
           if (aData.children) {
             aData.children.forEach(function(aChild) {
-              this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              try {
+                this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              } catch (ex) {
+                // invalid tag child, skip it
+              }
             }, this);
             return [folderIdMap, searchIds];
           }
@@ -1426,9 +1445,9 @@ var PlacesUtils = {
           var childNode = aSourceNode.getChild(i);
           if (aExcludeItems && aExcludeItems.indexOf(childNode.itemId) != -1)
             continue;
-          if (i != 0)
+          var written = serializeNodeToJSONStream(aSourceNode.getChild(i), i);
+          if (written && i < cc - 1)
             aStream.write(",", 1);
-          serializeNodeToJSONStream(aSourceNode.getChild(i), i);
         }
         if (!wasOpen)
           aSourceNode.containerOpen = false;
@@ -1449,26 +1468,44 @@ var PlacesUtils = {
 
       addGenericProperties(bNode, node);
 
+      var parent = bNode.parent;
+      var grandParent = parent ? parent.parent : null;
+
       if (self.nodeIsURI(bNode)) {
+        // Tag root accept only folder nodes
+        if (parent && parent.itemId == self.tagsFolderId)
+          return false;
         // Check for url validity, since we can't halt while writing a backup.
         // This will throw if we try to serialize an invalid url and it does
         // not make sense saving a wrong or corrupt uri node.
         try {
           self._uri(bNode.uri);
         } catch (ex) {
-          return;
+          return false;
         }
         addURIProperties(bNode, node);
       }
-      else if (self.nodeIsContainer(bNode))
+      else if (self.nodeIsContainer(bNode)) {
+        // Tag containers accept only uri nodes
+        if (grandParent && grandParent.itemId == self.tagsFolderId)
+          return false;
         addContainerProperties(bNode, node);
-      else if (self.nodeIsSeparator(bNode))
+      }
+      else if (self.nodeIsSeparator(bNode)) {
+        // Tag root accept only folder nodes
+        // Tag containers accept only uri nodes
+        if ((parent && parent.itemId == self.tagsFolderId) ||
+            (grandParent && grandParent.itemId == self.tagsFolderId))
+          return false;
+
         addSeparatorProperties(bNode, node);
+      }
 
       if (!node.feedURI && node.type == self.TYPE_X_MOZ_PLACE_CONTAINER)
         writeComplexNode(aStream, node, bNode);
       else
         writeScalarNode(aStream, node);
+      return true;
     }
 
     // serialize to stream
